/*
    Copyright (C) 2007  Miklos Szeredi <miklos@szeredi.hu>

    This program can be distributed under the terms of the GNU GPL.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <pwd.h>
#include <sys/mount.h>
#include <sys/fsuid.h>
#include <sys/capability.h>

#ifndef MS_DIRSYNC
#define MS_DIRSYNC	128
#endif
#ifndef MS_MOVE
#define MS_MOVE		8192
#endif
#ifndef MS_REC
#define MS_REC		16384
#endif
#ifndef MS_UNBINDABLE
#define MS_UNBINDABLE	(1<<17)
#endif
#ifndef MS_PRIVATE
#define MS_PRIVATE	(1<<18)
#endif
#ifndef MS_SLAVE
#define MS_SLAVE	(1<<19)
#endif
#ifndef MS_SHARED
#define MS_SHARED	(1<<20)
#endif
#ifndef MS_SETUSER
#define MS_SETUSER	(1<<22)
#endif

struct mount_flags {
    const char *opt;
    unsigned long flag;
    int on;
};

static struct mount_flags mount_flags[] = {
	{"rw",		MS_RDONLY,	0},
	{"ro",		MS_RDONLY,	1},
	{"suid",	MS_NOSUID,	0},
	{"nosuid",	MS_NOSUID,	1},
	{"dev",		MS_NODEV,	0},
	{"nodev",	MS_NODEV,	1},
	{"exec",	MS_NOEXEC,	0},
	{"noexec",	MS_NOEXEC,	1},
	{"async",	MS_SYNCHRONOUS,	0},
	{"sync",	MS_SYNCHRONOUS,	1},
	{"remount",	MS_REMOUNT,	1},
	{"mand",	MS_MANDLOCK,	1},
	{"nomand",	MS_MANDLOCK,	0},
	{"dirsync",	MS_DIRSYNC,	1},
	{"atime",	MS_NOATIME,	0},
	{"noatime",	MS_NOATIME,	1},
	{"nodiratime",	MS_NODIRATIME,	1},
	{NULL,		0,		0}
};

static int cmd = 0;
static struct option long_opts[] = {
	{"help",	0,	NULL,	'h'},
	{"version",	0,	NULL,	'V'},
	{"bind",	0,	&cmd,	MS_BIND},
	{"move",	0,	&cmd,	MS_MOVE},
	{"rec",		0,	NULL,	'R'},
	{"unbindable",	0,	&cmd,	MS_UNBINDABLE},
	{"private",	0,	&cmd,	MS_PRIVATE},
	{"slave",	0,	&cmd,	MS_SLAVE},
	{"shared",	0,	&cmd,	MS_SHARED},
	{NULL,		0,	NULL,	0}
};


static int find_mount_flag(const char *opt, int *flag)
{
	int i;

	for (i = 0; mount_flags[i].opt != NULL; i++) {
		if (strcmp(mount_flags[i].opt, opt) == 0) {
			if (mount_flags[i].on)
				*flag |= mount_flags[i].flag;
			else
				*flag &= ~mount_flags[i].flag;
			return 1;
		}
        }
	return 0;
}

static uid_t get_uid(const char *user)
{
	char *end;
	uid_t uid;
	struct passwd *p;

	if (!user[0]) {
		fprintf(stderr, "empty username\n");
		exit(1);
	}

	uid = strtol(user, &end, 0);
	if (!end[0])
		return uid;

	p = getpwnam(user);
	if (!p) {
		fprintf(stderr, "invalid user: %s\n", user);
		exit(1);
	}
	return p->pw_uid;
}

static char *append_opt(char *opts, const char *opt)
{
	size_t oldlen = opts ? strlen(opts) : 0;
	opts = realloc(opts, oldlen + strlen(opt) + 2);
	if (!opts) {
		fprintf(stderr, "Out of memory\n");
		exit(1);
	}
	if (oldlen)
		opts[oldlen++] = ',';
	strcpy(opts+oldlen, opt);
	return opts;
}

static void usage(const char *progname)
{
	fprintf(stderr, 
		"usage: %s [-t type] [-o opts] [misc_opts] [[source] target]\n"
		"misc_opts:\n"
		"  --help -h        help\n"
		"  --version -V     version\n"
		"  -r               read-only\n"
		"  -w               read-write\n"
		"  --rec -R         recursive\n"
		"  --bind           bind\n"
		"  --move           move\n"
		"  --shared         make shared\n"
		"  --slave          make slave\n"
		"  --private        make private\n"
		"  --unbindable     make unbindable\n"
		, progname);
}

int main(int argc, char *argv[])
{
	char *type = "";
	char *opts = NULL;
	char *user = NULL;
	char *source = "";
	char *target = "";
	char *opt;
	char *newopts;
	int flags = 0;
	int res;

	if (argc == 1) {
		usage(argv[0]);
		exit(1);
	}
        while (1) {
		int c = getopt_long(argc, argv, "t:o:rwRVh", long_opts, NULL);
		if (c == -1)
			break;

		switch (c) {
		case 0:
			flags = (flags & MS_REC) | cmd;
			break;

		case 't':
			type = optarg;
			break;

		case 'o':
			opts = append_opt(opts, optarg);
			break;

		case 'r':
			opts = append_opt(opts, "ro");
			break;

		case 'w':
			opts = append_opt(opts, "rw");
			break;

		case 'R':
			flags |= MS_REC;
			break;

		case 'V':
			fprintf(stderr, "version: %s\n", VERSION);
			exit(0);

		case 'h':
		case '?':
			usage(argv[0]);
			exit(1);
		}
	}

	newopts = NULL;
	while (1) {
		opt = strtok(opts, ",");
		if (!opt)
			break;

		if (!find_mount_flag(opt, &flags)) {
			if (strncmp(opt, "user=", 5) == 0)
				user = opt + 5;
			else
				newopts = append_opt(newopts, opt);
		}
		opts = NULL;
	}

	if (optind < argc)
		source = argv[optind++];
	if (optind < argc)
		target = argv[optind++];
	else {
		target = source;
		source = "";
	}

	if (user) {
		cap_t cap;
		uid_t uid = get_uid(user);
		int res;

		cap = cap_get_proc();
		if (!cap) {
			perror("failed to get capabilities");
			exit(1);
		}
		res = setfsuid(uid);
		if (res == -1) {
			perror("failed to set filesystem user id");
			exit(1);
		}
		res = cap_set_proc(cap);
		if (res == -1) {
			perror("failed to set capabilities");
		}

		flags |= MS_SETUSER;
	}

	res = mount(source, target, type, flags, newopts);
	if (res == -1) {
		perror(argv[0]);
		exit(1);
	}
	return 0;
}
